"""Localization utilities to translate error messages"""
# pylint: disable=line-too-long

import sys

translations = {
    # Error messages
    "INVALID_LENGTH": {
        "en": "Invalid length for parameter `{}`",
        "it": "Lunghezza non valida per il parametro `{}`",
        "de": "Ungültige Länge für den Parameter `{}`",
        "fr": "Longeur invalide pour paramètre `{}`",
    },
    "INVALID_TYPE": {
        "en": "Invalid type for parameter `{}`: expected `{}`, got `{}`",
        "it": "Tipo non valido per il parametro `{}`: previsto `{}`, ricevuto `{}`",
        "de": "Ungültiger Typ für den Parameter `{}`: erwartet `{}`, erhalten `{}`",
        "fr": "Type invalid pour le paramètre `{}`: attendu `{}`, eu `{}`"
    },
    "INVALID_ELEMENT_TYPE": {
        "en": "Invalid type for parameter `{0}`: at position {3} expected `{1}`, got `{2}`",
        "it": "Tipo non valido per il parametro `{0}`: alla posizione {3} previsto `{1}`, ricevuto `{2}`",
        "de": "Ungültiger Typ für den Parameter `{0}`: an Position {3} erwartet `{1}`, erhalten `{2}`",
        "fr": "Type invalide pour le paramètre `{0}`: à la position {3} attendu `{1}`, eu `{2}`",
    },
    "INVALID_RANGE": {
        "en": "Value for parameter `{}` out of the allowed range [{}, {}]",
        "it": "Valore per il parametro `{}` fuori dall'intervallo consentito [{}, {}]",
        "de": "Wert für den Parameter `{}` außerhalb des zulässigen Bereichs [{}, {}]",
        "fr": "Valeur pour le paramètre `{}` est hors de l'interval autorisé [{}, {}]",
    },
    "INVALID_FILENAME_EXTENSION": {
        "en": "Invalid or missing extension for filename: only .png and .svg are supported",
        "it": "Estensione non valida o mancante per il nome del file: sono supportati solo .png e .svg",
        "de": "Ungültige oder fehlende Dateinamenerweiterung: Nur .png und .svg werden unterstützt",
        "fr": "Extension non-valid ou mancante pour le nom du fichier: sont supportées seulement .png et .svg",
    },
    "INVALID_FILENAME_GIF": {
        "en": "Invalid or missing extension for a GIF file: needs to end with `.gif`",
        "it": "Estensione non valida o mancante per un file GIF: deve terminare con `.gif`",
        "de": "Ungültige oder fehlende Dateinamenerweiterung für eine GIF-Datei: muss mit `.gif` enden",
        "fr": "Extension invalide ou manquante pour un fichier GIF: doit terminer en `.gif`",
    },
    "EMPTY_GRAPHICS_LIST": {
        "en": "The list of graphics cannot be empty",
        "it": "La lista delle grafiche non può essere vuota",
        "de": "Die Liste der Grafiken darf nicht leer sein",
        "fr": "La list de graphiques ne peut pas être vide",
    },
    "EMPTY_AREA_OUTPUT": {
        "en": "Cannot show/save a graphic of size {} as it has no area",
        "it": "Impossibile mostrare/salvare una grafica di dimensione {} poiché non ha area",
        "de": "Kann eine Grafik der Größe {} nicht anzeigen/speichern, da sie keine Fläche hat",
        "fr": "Impossible de montrer/sauvegarder un graphiqye de dimensions {} vu qu'il ne possède pas de surface",
    },
    "FONT_NOT_FOUND": {
        "en": "Cannot find font '{}', another font used as fallback",
        "it": "Impossibile trovare il font '{}', un altro font è stato usato come alternativa",
        "de": "Schriftart '{}' nicht gefunden, eine andere Schriftart wird als Fallback verwendet",
        "fr": "Impossible de trouver la police '{}', une autre police est utilisée comme alternative",
    },
    "DIFFERENT_SIZES": {
        "en": "All graphics in the list must have the same size",
        "it": "Tutte le grafiche nella lista devono avere le stesse dimensioni",
        "de": "Alle Grafiken in der Liste müssen die gleiche Größe haben",
        "fr": "Tous les graphiques dans la liste doivent avoir la même taille",
    },
    # Function names
    "compose": {
        "en": "compose",
        "it": "componi",
        "de": "kombiniere",
        "fr": "compose",
    },
    "pin": {
        "en": "pin",
        "it": "fissa",
        "de": "fixiere",
        "fr": "ancre",
    },
    "overlay": {
        "en": "overlay",
        "it": "sovrapponi",
        "de": "ueberlagere",
        "fr": "superpose",
    },
    "beside": {
        "en": "beside",
        "it": "accanto",
        "de": "neben",
        "fr": "cote_a_cote",
    },
    "above": {
        "en": "above",
        "it": "sopra",
        "de": "ueber",
        "fr": "au_dessus",
    },
    "rotate": {
        "en": "rotate",
        "it": "ruota",
        "de": "drehe",
        "fr": "pivote",
    },
    "rgb_color": {
        "en": "rgb_color",
        "it": "colore_rgb",
        "de": "rgb_farbe",
        "fr": "couleur_rgb",
    },
    "rectangle": {
        "en": "rectangle",
        "it": "rettangolo",
        "de": "rechteck",
        "fr": "rectangle",
    },
    "empty_graphic": {
        "en": "empty_graphic",
        "it": "grafica_vuota",
        "de": "leere_grafik",
        "fr": "graphique_vide",
    },
    "ellipse": {
        "en": "ellipse",
        "it": "ellisse",
        "de": "ellipse",
        "fr": "ellipse",
    },
    "text": {
        "en": "text",
        "it": "testo",
        "de": "text",
        "fr": "text",
    },
    "circular_sector": {
        "en": "circular_sector",
        "it": "settore_circolare",
        "de": "kreissektor",
        "fr": "secteur_circulaire",
    },
    "triangle": {
        "en": "triangle",
        "it": "triangolo",
        "de": "dreieck",
        "fr": "triangle",
    },
    # Parameter names
    "width": {
        "en": "width",
        "it": "larghezza",
        "de": "breite",
        "fr": "largeur",
    },
    "height": {
        "en": "height",
        "it": "altezza",
        "de": "hoehe",
        "fr": "hauteur",
    },
    "radius": {
        "en": "radius",
        "it": "raggio",
        "de": "radius",
        "fr": "rayon",
    },
    "opacity": {
        "en": "opacity",
        "it": "opacita",
        "de": "opazitaet",
        "fr": "opacite",
    },
    "hue": {
        "en": "hue",
        "it": "tonalita",
        "de": "farbton",
        "fr": "teinte",
    },
    "saturation": {
        "en": "saturation",
        "it": "saturazione",
        "de": "saettigung",
        "fr": "saturation",
    },
    "value": {
        "en": "value",
        "it": "valore",
        "de": "hellwert",
        "fr": "valeur",
    },
    "lightness": {
        "en": "lightness",
        "it": "luce",
        "de": "helligkeit",
        "fr": "luminosite",
    },
    "angle": {
        "en": "angle",
        "it": "angolo",
        "de": "winkel",
        "fr": "angle",
    },
    "side1": {
        "en": "side1",
        "it": "lato1",
        "de": "seite1",
        "fr": "cote1",
    },
    "side2": {
        "en": "side2",
        "it": "lato2",
        "de": "seite2",
        "fr": "cote2",
    },
    "content": {
        "en": "content",
        "it": "contenuto",
        "de": "inhalt",
        "fr": "contenu",
    },
    "foreground_graphic": {
        "en": "foreground_graphic",
        "it": "grafica_primopiano",
        "de": "vordere_grafik",
        "fr": "graphique_premier_plan",
    },
    "background_graphic": {
        "en": "background_graphic",
        "it": "grafica_secondopiano",
        "de": "hintere_grafik",
        "fr": "graphique_arriere_plan",
    },
    "left_graphic": {
        "en": "left_graphic",
        "it": "grafica_sinistra",
        "de": "linke_grafik",
        "fr": "graphique_gauche",
    },
    "right_graphic": {
        "en": "right_graphic",
        "it": "grafica_destra",
        "de": "rechte_grafik",
        "fr": "graphique_droite",
    },
    "top_graphic": {
        "en": "top_graphic",
        "it": "grafica_alto",
        "de": "obere_grafik",
        "fr": "graphique_haut",
    },
    "bottom_graphic": {
        "en": "bottom_graphic",
        "it": "grafica_basso",
        "de": "untere_grafik",
        "fr": "graphique_bas",
    },
    "filename": {
        "en": "filename",
        "it": "nome_file",
        "de": "datei_name",
        "fr": "nom_fichier",
    },
    "graphics": {
        "en": "graphics",
        "it": "grafiche",
        "de": "grafiken",
        "fr": "graphiques",
    },
    "graphic": {
        "en": "graphic",
        "it": "grafica",
        "de": "grafiken",
        "fr": "graphique",
    },
    "color": {
        "en": "color",
        "it": "colore",
        "de": "farbe",
        "fr": "couleur",
    },
    "point": {
        "en": "point",
        "it": "punto",
        "de": "punkt",
        "fr": "point",
    },
    "font": {
        "en": "font",
        "it": "font",
        "de": "schriftart",
        "fr": "police",
    },
    "points": {
        "en": "points",
        "it": "punti",
        "de": "punkte",
        "fr": "points",
    },
    "debug": {
        "en": "debug",
        "it": "debug",
        "de": "debug",
        "fr": "debug",
    },
    "duration": {
        "en": "duration",
        "it": "durata",
        "de": "dauer",
        "fr": "duree",
    },
    "loop": {
        "en": "loop",
        "it": "loop",
        "de": "loop",
        "fr": "en_boucle",
    },
    # Types
    "Graphic": {
        "en": "Graphic",
        "it": "Grafica",
        "de": "Grafik",
        "fr": "Graphique",
    },
    "Color": {
        "en": "Color",
        "it": "Colore",
        "de": "Farbe",
        "fr": "Couleur",
    },
    "Point": {
        "en": "Point",
        "it": "Punto",
        "de": "Punkt",
        "fr": "Point",
    },
    # Known Points
    "center": {
        "en": "center",
        "it": "centro",
        "de": "mitte",
        "fr": "centre",
    },
    "top_center": {
        "en": "top_center",
        "it": "alto_centro",
        "de": "oben_mitte",
        "fr": "haut_centre",
    },
    "bottom_center": {
        "en": "bottom_center",
        "it": "basso_centro",
        "de": "unten_mitte",
        "fr": "bas_centre",
    },
    "center_left": {
        "en": "center_left",
        "it": "centro_sinistra",
        "de": "mitte_links",
        "fr": "centre_gauche",
    },
    "center_right": {
        "en": "center_right",
        "it": "centro_destra",
        "de": "mitte_rechts",
        "fr": "centre_droite",
    },
    "top_left": {
        "en": "top_left",
        "it": "alto_sinistra",
        "de": "oben_links",
        "fr": "haut_gauche",
    },
    "top_right": {
        "en": "top_right",
        "it": "alto_destra",
        "de": "oben_rechts",
        "fr": "haut_droite",
    },
    "bottom_left": {
        "en": "bottom_left",
        "it": "basso_sinistra",
        "de": "unten_links",
        "fr": "bas_gauche",
    },
    "bottom_right": {
        "en": "bottom_right",
        "it": "basso_destra",
        "de": "unten_rechts",
        "fr": "bas_droite",
    },
    # Known Colors
    "black": {
        "en": "black",
        "it": "nero",
        "de": "schwarz",
        "fr": "noir",
    },
    "red": {
        "en": "red",
        "it": "rosso",
        "de": "rot",
        "fr": "rouge",
    },
    "green": {
        "en": "green",
        "it": "verde",
        "de": "gruen",
        "fr": "vert",
    },
    "blue": {
        "en": "blue",
        "it": "blu",
        "de": "blau",
        "fr": "bleu",
    },
    "yellow": {
        "en": "yellow",
        "it": "giallo",
        "de": "gelb",
        "fr": "jaune",
    },
    "magenta": {
        "en": "magenta",
        "it": "magenta",
        "de": "magenta",
        "fr": "magenta",
    },
    "cyan": {
        "en": "cyan",
        "it": "ciano",
        "de": "cyan",
        "fr": "cyan",
    },
    "white": {
        "en": "white",
        "it": "bianco",
        "de": "weiss",
        "fr": "blanc",
    },
    "transparent": {
        "en": "transparent",
        "it": "trasparente",
        "de": "transparent",
        "fr": "transparent",
    },
}


def translate(key: str, *args) -> str:
    """Looks up an error message by `key` and returns the localized version.

    :param key: error message key
    :param args: optional arguments to be formatted into the localized message
    :returns: locale-specific descriptive error message
    """
    lang = sys.modules["pytamaro"].LANGUAGE
    if key not in translations:
        return key
    localized = translations[key][lang]
    if len(localized) == 0:
        localized = translations[key]["en"]
    return localized.format(*args)



"""
Fonctions pour faire des opérations sur des graphiques (principalement pour les combiner).
"""

from __future__ import annotations

from pytamaro import operations
from pytamaro.fr.graphic import Graphique
from pytamaro.fr.point import Point
from pytamaro.operations import (above, beside, graphic_height, graphic_width,
                                 overlay, pin, rotate)


def largeur_graphique(graphique: Graphique) -> int:
    """
    Retourne la largeur du graphique.

    :param graphique: graphique duquel calculer la largeur
    :returns: la largeur du graphique
    """
    return graphic_width(graphique)


def hauteur_graphique(graphique: Graphique) -> int:
    """
    Retourne la hauteur du graphique.

    :param graphique: graphique duquel calculer la hauteur
    :returns: la hauteur du graphique
    """
    return graphic_height(graphique)


def compose(graphique_premier_plan: Graphique, graphique_arriere_plan: Graphique) \
        -> Graphique:
    """
    Crée un nouveau graphique en composant les deux graphiques fournis.
    Le premier graphique est maintenu au premier plan, le second en arrière plan.
    Les graphiques sont alignés en superposant les points d'ancrage.

    Le point d'ancrage du graphique résultant est celui utilisé pour la composition.

    :param graphique_premier_plan: le graphique au premier plan
    :param graphique_premier_plan: le graphique en arrière plan
    :returns: le graphique qui résulte de la composition
    """
    return operations.compose(graphique_premier_plan, graphique_arriere_plan)


def ancre(point: Point, graphique: Graphique) -> Graphique:
    """
    Créé un nouveau graphique qui correspond au graphique fourni avec un nouveau point d'ancrage.

    Chaque graphique est compris dans un cadre de délimitation.
    Il y a 9 points d'ancrages particuliers, qui correspondent aux quatre coins du cadre
    (rectangle), le milieu de chaque côté ainsi que le centre du cadre.
    Les points peuvent être désignés avec les noms suivants: `haut_gauche`, `haut_droite`,
    `bas_gauche`, `bas_droite`, `haut_centre`, `centre_droite`, `bas_centre`, `centre_gauche` et
    `centre`.

    :param point: le point indiquant le nouveau point d'ancrage
    :param graphique: le graphique originel
    :returns: un nouveau graphique avec le nouveau point d'ancrage
    """
    return pin(point, graphique)


def superpose(graphique_premier_plan: Graphique, graphique_arriere_plan: Graphique) \
        -> Graphique:
    """
    Créé un nouveau graphique en superposant les deux graphiques fournis, en gardant le premier au
    premier plan et en mettant le second en arrière plan.
    Les graphiques sont superposés par rapport à leur centre.

    Le point d'ancrage du nouveau graphique est en son centre.

    :param graphique_premier_plan: le graphique au premier plan
    :param graphique_arriere_plan: le graphique en arrière plan
    :returns: le graphique résultant de la superposition des deux graphiques fournis
    """
    return overlay(graphique_premier_plan, graphique_arriere_plan)


def cote_a_cote(graphique_gauche: Graphique, graphique_droite: Graphique) -> Graphique:
    """
    Créé un graphique en plaçant les deux graphiques fournis côte à côte.
    Les deux graphiques sont centrés verticalement.

    Le point d'ancrage du nouveau graphique est en son centre.

    :param graphique_gauche: le graphique à placer à gauche
    :param graphique_droite: le graphique à placer à droite
    :returns: le graphique résultant après avoir placé les deux graphiques l'un à côté de l'autre
    """
    return beside(graphique_gauche, graphique_droite)


def au_dessus(graphique_haut: Graphique, graphique_bas: Graphique) -> Graphique:
    """
    Créé un graphique en plaçant les deux graphiques fournis l'un au-dessus de l'autre.
    Les deux graphiques sont centrés horizontalement.

    Le point d'ancrage du nouveau graphique est en son centre.

    :param graphique_haut: le graphique à placer au dessus
    :param graphique_bas: le graphique à placer en dessous
    :returns: le graphique résultant après avoir placé les deux graphiques l'un au-dessus de l'autre
    """
    return above(graphique_haut, graphique_bas)


def pivote(angle: float, graphique: Graphique) -> Graphique:
    """
    Crée un nouveau graphique en pivotant dans le sens inverse des aiguilles d'une montre le
    graphique fourni autour de son point d'ancrage selon l'angle donné. Un angle négatif
    correspond à une rotation dans le sens des aiguilles d'une montre.

    :param angle: angle dans le sens inverse des aiguilles d'une montre, en degrés
    :param graphique: le graphique à pivoter
    :returns: un nouveau graphique pivoté
    """
    return rotate(angle, graphique)



"""
Noms de couleurs particulières qui sont aux coins du cube RGB.
Ainsi que la couleur complètement transparent.
"""

from __future__ import annotations

from pytamaro.color_names import black, blue, cyan, green, magenta, red, transparent, white, yellow
from pytamaro.fr.color import Couleur

noir: Couleur = black
"""Couleur noire

:meta hide-value:
"""

rouge: Couleur = red
"""Couleur rouge

:meta hide-value:
"""

vert: Couleur = green
"""Couleur verte

:meta hide-value:
"""

bleu: Couleur = blue
"""Couleur bleue

:meta hide-value:
"""

jaune: Couleur = yellow
"""Couleur jaune

:meta hide-value:
"""

magenta: Couleur = magenta  # pylint: disable=self-assigning-variable
"""Couleur magenta

:meta hide-value:
"""

cyan: Couleur = cyan
"""Couleur cyan

:meta hide-value:
"""

blanc: Couleur = white
"""Couleur blanche

:meta hide-value:
"""

transparent: Couleur = transparent  # pylint: disable=self-assigning-variable
"""Couleur complètement transparente

:meta hide-value:
"""



"""
Type `Graphic`, that includes a graphic with a pinning position.
"""

import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property

from skia import (Canvas, Font, FontMgr, Matrix, Paint, Path, Point, Rect,
                  Size, Typeface)

from pytamaro.color import Color
from pytamaro.localization import translate
from pytamaro.point import Point as PyTamaroPoint
from pytamaro.point_names import (bottom_center, center, center_left,
                                  center_right, top_center)


@dataclass(frozen=True, eq=False)
class Graphic(ABC):
    """
    A graphic (image) with a position for pinning.

    The pinning position is used in the following operations:

    - rotation (to determine the center of rotation)
    - graphic composition (two graphics get composed aligning their pinning
      position).
    """
    pin_position: Point
    path: Path

    def size(self) -> Size:
        """
        Computes the size of this graphic (x and y axes spanning),
        using the bounds computed by bounds().

        :returns: graphic's size
        """
        return Size(self.bounds.width(), self.bounds.height())

    @cached_property
    def bounds(self) -> Rect:
        """
        Computes the (tight) bounds for the path (outline) of this graphic.

        :returns: a rectangle that indicates the bounds of the graphic in the 2D
                  space
        """
        return self.path.computeTightBounds()

    @abstractmethod
    def draw(self, canvas: Canvas):
        """
        Draws the current graphic onto the provided canvas.

        :param canvas: canvas onto which to draw
        """

    def zero_pixels(self) -> bool:
        """
        Returns whether this graphic has no pixels to render, because its (rounded) area is 0.

        :returns: True if the graphic has no pixels, False otherwise
        """
        return self.size().toRound().isEmpty()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Graphic):
            return self.__hash__() == other.__hash__()
        return False

    def __hash__(self) -> int:
        return hash(self._key())

    def _key(self):
        return ((self.pin_position.fX, self.pin_position.fY),
                self.path.serialize().bytes())


@dataclass(frozen=True, eq=False)
class Primitive(Graphic):
    """
    Represents a primitive graphic, which has a uniform color.
    Geometric shapes and text are primitive graphics.
    """
    color: Color

    def __init__(self, path: Path, color: Color, pin_position: Point = None):
        object.__setattr__(self, "color", color)
        if pin_position is None:
            bounds = path.computeTightBounds()
            pin_position = Point(bounds.width() / 2, bounds.height() / 2)
        super().__init__(pin_position, path)

    @cached_property
    def fill_paint(self) -> Paint:
        """
        Paint used to fill the graphic.
        """
        return Paint(Color=self.color.skia_color, AntiAlias=False, Style=Paint.kFill_Style)

    @cached_property
    def stroke_paint(self) -> Paint:
        """
        Paint used to draw the outline of the graphic.
        """
        return Paint(Color=self.color.skia_color, AntiAlias=True, Style=Paint.kStroke_Style)

    def draw(self, canvas: Canvas):
        canvas.drawPath(self.path, self.fill_paint)
        canvas.drawPath(self.path, self.stroke_paint)

    def _key(self):
        return super()._key(), self.color


@dataclass(frozen=True, eq=False)
class Empty(Graphic):
    """
    An empty graphic.
    """

    def __init__(self):
        super().__init__(Point(0, 0), Path())

    def draw(self, canvas: Canvas):
        pass

    def __repr__(self) -> str:
        return f"{translate('empty_graphic')}()"


@dataclass(frozen=True, eq=False)
class Rectangle(Primitive):
    """
    A rectangle.
    """
    width: float
    height: float

    def __init__(self, width: float, height: float, color: Color):
        object.__setattr__(self, "width", width)
        object.__setattr__(self, "height", height)
        path = Path().addRect(Rect.MakeWH(width, height))
        super().__init__(path, color)

    def __repr__(self) -> str:
        return f"{translate('rectangle')}({self.width}, {self.height}, {self.color})"


@dataclass(frozen=True, eq=False)
class Ellipse(Primitive):
    """
    An ellipse.
    """
    width: float
    height: float

    def __init__(self, width: float, height: float, color: Color):
        object.__setattr__(self, "width", width)
        object.__setattr__(self, "height", height)
        path = Path().addOval(Rect.MakeWH(width, height))
        super().__init__(path, color)

    def __repr__(self) -> str:
        return f"{translate('ellipse')}({self.width}, {self.height}, {self.color})"


@dataclass(frozen=True, eq=False)
class CircularSector(Primitive):
    """
    A circular sector (with an angle between 0 and 360).
    Its pinning position is the center of the circle from which it is taken.
    """
    radius: float
    angle: float

    def __init__(self, radius: float, angle: float, color: Color):
        object.__setattr__(self, "radius", radius)
        object.__setattr__(self, "angle", angle)
        if angle == 360:
            path = Path.Circle(radius, radius, radius)
        else:
            diameter = 2 * radius
            path = Path()
            path.moveTo(radius, radius)
            path.arcTo(Rect.MakeWH(diameter, diameter), 0, -angle, False)
            path.close()
        super().__init__(path, color, Point(radius, radius))

    def __repr__(self) -> str:
        return f"{translate('circular_sector')}({self.radius}, {self.angle}, {self.color})"


@dataclass(frozen=True, eq=False)
class Triangle(Primitive):
    """
    A triangle specified using two sides and the angle between them.
    The first side extends horizontally to the right.
    The second side is rotated counterclockwise by the specified angle.

    Its pinning position is the centroid of the triangle.
    """
    side1: float
    side2: float
    angle: float

    def __init__(self, side1: float, side2: float, angle: float, color: Color):
        object.__setattr__(self, "side1", side1)
        object.__setattr__(self, "side2", side2)
        object.__setattr__(self, "angle", angle)
        third_point = Matrix.RotateDeg(-angle).mapXY(side2, 0)
        path = Path.Polygon([Point(0, 0), Point(side1, 0), third_point], isClosed=True)
        # The centroid is the average of the three vertices
        centroid = Point((side1 + third_point.x()) / 3, third_point.y() / 3)
        super().__init__(path, color, centroid)

    def __repr__(self) -> str:
        return f"{translate('triangle')}({self.side1}, {self.side2}, {self.angle}, {self.color})"


@dataclass(frozen=True, eq=False)
class Text(Primitive):
    """
    Graphic containing text, using a given font with a given typographic size.
    Its pinning position is horizontally aligned on the left and vertically on
    the baseline of the text.
    """
    text: str
    font_name: str
    text_size: float

    def __init__(self, text: str, font_name: str, text_size: float, color: Color):
        object.__setattr__(self, "text", text)
        object.__setattr__(self, "font_name", font_name)
        object.__setattr__(self, "text_size", text_size)
        if FontMgr().matchFamily(font_name).count() == 0:
            print(translate("FONT_NOT_FOUND", font_name), file=sys.stderr)
        glyphs = self.font.textToGlyphs(text)
        offsets = self.font.getXPos(glyphs)
        text_path = Path()
        for glyph, x_offset in zip(glyphs, offsets):
            path = self.font.getPath(glyph)
            if path is not None:  # some glyphs (e.g., a space) have no outline
                path.offset(x_offset, 0)
                text_path.addPath(path)
        # The pinning position is on the left (0) on the baseline (0).
        super().__init__(text_path, color, Point(0, 0))

    @cached_property
    def font(self) -> Font:
        """
        The Skia Font used to render the text.
        """
        return Font(Typeface(self.font_name), self.text_size)

    @cached_property
    def bounds(self) -> Rect:
        """
        Computes the bounding box of the text, whose width is determined by
        Font.measureText() to account for leading and trailing glyphs with no outline.
        """
        path_bounds = super().bounds
        text_length = self.font.measureText(self.text)
        return Rect.MakeLTRB(0, path_bounds.top(), text_length, path_bounds.bottom())

    def __repr__(self) -> str:
        return f"{translate('text')}({self.text!r}, {self.font_name!r}, {self.text_size}, {self.color})"  # pylint: disable=line-too-long


@dataclass(frozen=True, eq=False)
class Compose(Graphic):
    """
    Represents the composition of two graphics, one in the foreground and the
    other in the background, joined on their pinning positions.
    """
    foreground: Graphic
    background: Graphic

    def __init__(self, foreground: Graphic, background: Graphic):
        object.__setattr__(self, "foreground", foreground)
        object.__setattr__(self, "background", background)
        fg_pin = self.foreground.pin_position
        bg_pin = self.background.pin_position
        pin = Point(bg_pin.x(), bg_pin.y())
        path = Path(self.background.path)
        path.addPath(self.foreground.path,
                     bg_pin.x() - fg_pin.x(), bg_pin.y() - fg_pin.y())
        super().__init__(pin, path)

    def draw(self, canvas: Canvas):
        canvas.save()
        self.background.draw(canvas)
        canvas.translate(self.background.pin_position.x() - self.foreground.pin_position.x(),
                         self.background.pin_position.y() - self.foreground.pin_position.y())
        self.foreground.draw(canvas)
        canvas.restore()

    def __repr__(self) -> str:
        return f"{translate('compose')}({self.foreground}, {self.background})"


@dataclass(frozen=True, eq=False)
class Pin(Graphic):
    """
    Represents the pinning of a graphic in a certain position on its bounds.
    """
    graphic: Graphic
    pinning_point: PyTamaroPoint

    def __init__(self, graphic: Graphic, pinning_point: PyTamaroPoint):
        object.__setattr__(self, "graphic", graphic)
        object.__setattr__(self, "pinning_point", pinning_point)
        bounds = graphic.bounds
        h_mapping = {
            -1.0: bounds.left(),
            0.0: bounds.centerX(),
            1.0: bounds.right()
        }
        v_mapping = {
            1.0: bounds.top(),
            0.0: bounds.centerY(),
            -1.0: bounds.bottom()
        }
        pin = Point(h_mapping[pinning_point.x], v_mapping[pinning_point.y])
        super().__init__(pin, graphic.path)

    def draw(self, canvas: Canvas):
        self.graphic.draw(canvas)

    def __repr__(self) -> str:
        return f"{translate('pin')}({self.pinning_point}, {self.graphic})"


@dataclass(frozen=True, eq=False)
class Rotate(Graphic):
    """
    Represents the counterclockwise rotation of a graphic
    by a certain angle around the pinning position.
    The angle is expressed in degrees.
    """
    graphic: Graphic
    angle: float

    def __init__(self, graphic: Graphic, angle: float):
        object.__setattr__(self, "graphic", graphic)
        object.__setattr__(self, "angle", angle)
        # Negated angle because RotateDeg works clockwise.
        object.__setattr__(self, "rot_matrix", Matrix.RotateDeg(-angle, graphic.pin_position))
        path = Path()
        # transform() mutates the path provided as the second argument
        graphic.path.transform(self.rot_matrix, path)  # type: ignore  # pylint: disable=no-member
        super().__init__(graphic.pin_position, path)

    def draw(self, canvas: Canvas):
        canvas.save()
        canvas.concat(self.rot_matrix)  # type: ignore  # pylint: disable=no-member
        self.graphic.draw(canvas)
        canvas.restore()

    def __repr__(self) -> str:
        return f"{translate('rotate')}({self.angle}, {self.graphic})"


@dataclass(frozen=True, eq=False)
class SimpleCompose(Graphic):
    """
    Represents a simple composition operation between two graphics
    (i.e., beside, above, or overlay).
    These simple compositions pin the two graphics appropriately,
    compose them normally, and then pin the result on its center.
    """

    def __init__(self, graphic1: Graphic, graphic2: Graphic,
                 point1: PyTamaroPoint, point2: PyTamaroPoint):
        composed_graphic = Pin(Compose(Pin(graphic1, point1),
                                       Pin(graphic2, point2)), center)
        object.__setattr__(self, "composed_graphic", composed_graphic)
        super().__init__(composed_graphic.pin_position, composed_graphic.path)

    def draw(self, canvas: Canvas):
        self.composed_graphic.draw(canvas)  # type: ignore  # pylint: disable=no-member


@dataclass(frozen=True, eq=False)
class Beside(SimpleCompose):
    """
    Represents the composition of two graphics one beside the other,
    vertically centered.
    """
    left_graphic: Graphic
    right_graphic: Graphic

    def __init__(self, left_graphic: Graphic, right_graphic: Graphic):
        object.__setattr__(self, "left_graphic", left_graphic)
        object.__setattr__(self, "right_graphic", right_graphic)
        super().__init__(left_graphic, right_graphic, center_right, center_left)

    def __repr__(self) -> str:
        return f"{translate('beside')}({self.left_graphic}, {self.right_graphic})"


@dataclass(frozen=True, eq=False)
class Above(SimpleCompose):
    """
    Represents the composition of two graphics one above the other,
    horizontally centered.
    """
    top_graphic: Graphic
    bottom_graphic: Graphic

    def __init__(self, top_graphic: Graphic, bottom_graphic: Graphic):
        object.__setattr__(self, "top_graphic", top_graphic)
        object.__setattr__(self, "bottom_graphic", bottom_graphic)
        super().__init__(top_graphic, bottom_graphic, bottom_center, top_center)

    def __repr__(self) -> str:
        return f"{translate('above')}({self.top_graphic}, {self.bottom_graphic})"


@dataclass(frozen=True, eq=False)
class Overlay(SimpleCompose):
    """
    Represents the composition of two graphics that one overlay other,
    the center of the two graphics are at the same position
    """
    front_graphic: Graphic
    back_graphic: Graphic

    def __init__(self, front_graphic: Graphic, back_graphic: Graphic):
        object.__setattr__(self, "front_graphic", front_graphic)
        object.__setattr__(self, "back_graphic", back_graphic)
        super().__init__(front_graphic, back_graphic, center, center)

    def __repr__(self) -> str:
        return f"{translate('overlay')}({self.front_graphic}, {self.back_graphic})"



"""
Fonctions pour visualiser ou sauvegarder des graphiques.
"""

from __future__ import annotations

from pytamaro.io import save_animation, save_graphic, show_animation, show_graphic
from pytamaro.fr.graphic import Graphique


def montre_graphique(graphique: Graphique, debug: bool = False):
    """
    Montre un graphique. Les graphiques n'ayant pas de surface ne peuvent pas
    être montrés.

    Quand `debug` est `True`, la visualisation est ornée d'informations utiles
    pour le débogage:
    une bordure rouge autour du cadre de délimitation et une croix jaunâtre
    autour du point d'ancrage.

    :param graphique: le graphique à montrer
    :param debug: (optionnel) peut être assigné à `True` pour superposer les
           informations de débogage
    """
    show_graphic(graphique, debug)


def sauvegarde_graphique(nom_fichier: str, graphique: Graphique, debug: bool = False):
    """
    Sauvegarde un graphique dans un fichier.
    Deux formats de fichiers sont supportés: PNG (graphiques tramés) et SVG
    (graphiques vectoriels). L'extension du fichier (".png" ou ".svg") détermine
    le format.


    Les graphiques sans surface ne peuvent pas être sauvegardés au format
    PNG.

    Quand `debug` est `True`, la visualisation est ornée d'informations utiles
    pour le débogage:
    une bordure rouge autour du cadre de délimitation et une croix jaunâtre
    autour du point d'ancrage.

    :param nom_fichier: le nom du fichier à créer (avec l'extension)
    :param graphique: le graphique à sauvegarder
    :param debug: (optionnel) peut être assigné à `True` pour superposer les
           informations de débogage
    """
    save_graphic(nom_fichier, graphique, debug)


def sauvegarde_animation(
    nom_fichier: str, graphiques: list[Graphique], duree: int = 40, en_boucle: bool = True
):
    """
    Sauvegarde une séquence de graphiques en tant qu'animation (GIF).

    Les graphiques sont reproduits de manière séquentielle (normalement à 25
    images par secondes) en boucle (à moins que ça ne soit indiqué autrement).


    :param nom_fichier: le nom du fichier à créer (contenant l'extension ".gif")
    :param graphiques: la liste des graphiques à sauvegarder en tant
           qu'animation.
    :param duree: durée en millisecondes entre chaque image (par défaut est
           égale à 40 millisecondes, ce qui amène à avoir 25 images par secondes)
    :param en_boucle: si le GIF doit tourner en boucle indéfiniment (par défaut
           est `True`)
    """
    save_animation(nom_fichier, graphiques, duree, en_boucle)


def montre_animation(graphiques: list[Graphique], duree: int = 40, en_boucle: bool = True):
    """
    Montre une séquence de graphiques en tant qu'animation (GIF).

    Les graphiques sont reproduits de manière séquentielle (normalement à 25
    images par secondes) en boucle (à moins que ça ne soit indiqué autrement).


    :param graphiques: la liste des graphiques à sauvegarder en tant
           qu'animation.
    :param duree: durée en millisecondes entre chaque image (par défaut est
           égale à 40 millisecondes, ce qui amène à avoir 25 images par secondes)
    :param en_boucle: si le GIF doit tourner en boucle indéfiniment (par défaut
           est `True`)
    """
    show_animation(graphiques, duree, en_boucle)



"""
Type `Couleur`, ainsi des fonctions pour produire des couleurs et constantes représentants des
couleurs particulières.
"""

from __future__ import annotations

from pytamaro.color import Color
from pytamaro.color_functions import rgb_color, hsl_color, hsv_color

Couleur = Color
"""
Représente une couleur.
Une couleur a aussi un taux d'opacité,
de complètement transparent (comme la couleur `transparent`)
à complètement opaque (comme la couleur `rouge`).
"""


def couleur_rgb(rouge: int, vert: int, bleu: int, opacite: float = 1.0) -> Couleur:
    """
    Retourne une couleur avec les composantes indiquant le rouge (R), vert (G) et bleu (B)
    et un certain taux d'opacité (alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/thumb/b/b4/RGBCube_a.svg/524px-RGBCube_a.svg.png
       :height: 120px
       :align: center

       `Cube RGB (SharkD via Wikimedia Commons) <https://commons.wikimedia.org/wiki/File:RGBCube_a.svg>`_

    :param rouge: composante rouge [0-255]
    :param vert: composante verte [0-255]
    :param bleu: composante bleue [0-255]
    :param opacite: opacité (alpha) de la couleur, où 0 est complètement transparent
            et 1 complètement opaque.
            Par défaut, toutes les couleurs sont complètement opaques.
    :returns: une couleur avec les composantes RGBA données
    """
    return rgb_color(rouge, vert, bleu, opacite)


def couleur_hsv(teinte: float, saturation: float, valeur: float, opacite: float = 1.0) -> Couleur:
    """
    Retourne une couleur avec la teinte (H), saturation (S) et valeur (V) données,
    ainsi que le taux d'opacité (alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/4/4e/HSV_color_solid_cylinder.png
       :height: 120px
       :align: center

       `Cylindre HSV (SharkD via Wikimedia Commons) <https://commons.wikimedia.org/wiki/File:HSV_color_solid_cylinder.png>`_

    :param teinte: teinte de la couleur [0-360]
    :param saturation: saturation de la couleur [0-1]
    :param valeur: quantité de lumière appliquée [0-1]
           Les couleurs complètement saturées ont une quantité de lumière de 1.
    :param opacite: opacité (alpha) de la couleur, où 0 est complètement transparent
            et 1 complètement opaque.
            Par défaut, toutes les couleurs sont complètement opaques.
    :returns: une couleur avec le composantes HSVA données
    """
    return hsv_color(teinte, saturation, valeur, opacite)


def couleur_hsl(teinte: float, saturation: float, luminosite: float, opacite: float = 1.0
                ) -> Couleur:
    """
    Retourne une couleur avec la teinte (H), saturation (S) et luminosité (V) données,
    ainsi que le taux d'opacité (alpha, A).

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/3/35/HSL_color_solid_cylinder.png
       :height: 120px
       :align: center

       `Cylindre HSL: SharkD via Wikimedia Commons <https://commons.wikimedia.org/wiki/File:HSL_color_solid_cylinder.png>`_

    :param teinte: teinte de la couleur [0-360]
    :param saturation: saturation de la couleur [0-1]
    :param luminosite: quantité de blanc ou noire appliquée [0-1].
           Les couleurs complètement saturées ont un valeur de luminosité de 1/2.
    :param opacite: opacité (alpha) de la couleur, où 0 est complètement transparent
            et 1 complètement opaque.
            Par défaut, toutes les couleurs sont complètement opaques.
    :returns: une couleur avec les composantes HSLA données.
    """
    return hsl_color(teinte, saturation, luminosite, opacite)



# pylint:disable=duplicate-code
"""
Fonctions pour créer des graphiques primitifs (formes et texte).
À moins d'être spécifié autrement, le point d'ancrage initial est le centre du
cadre de délimitation (bounding box) du graphique.
"""

from __future__ import annotations

from pytamaro.fr.color import Couleur
from pytamaro.fr.graphic import Graphique
from pytamaro.primitives import circular_sector
from pytamaro.primitives import ellipse as en_ellipse
from pytamaro.primitives import empty_graphic
from pytamaro.primitives import rectangle as en_rectangle
from pytamaro.primitives import text
from pytamaro.primitives import triangle as en_triangle


def rectangle(largeur: float, hauteur: float, couleur: Couleur) -> Graphique:
    """
    Créé un rectangle de dimensions indiquées et rempli avec la couleur donnée.

    :param largeur: largeur du rectangle
    :param hauteur: hauteur du rectangle
    :param couleur: couleur qui remplira le rectangle
    :returns: un graphique avec le rectangle spécifié
    """
    return en_rectangle(largeur, hauteur, couleur)


def graphique_vide() -> Graphique:
    """
    Créé un graphique vide.
    Quand un graphique vide est composé avec n'importe quel autre graphique, il
    se comporte comme l'élément neutre: le résultat est toujours égal à l'autre
    graphique.

    :returns: un graphique vide (largeur et hauteur 0)
    """
    return empty_graphic()


def ellipse(largeur: float, hauteur: float, couleur: Couleur) -> Graphique:
    """
    Créé une ellipse avec les dimensions indiquées et rempli avec la couleur
    donnée.

    Lorsque la largeur et la hauteur coïncident, l'ellipse devient un cercle
    dont le diamètre est égal à la taille indiquée.

    :param largeur: largeur de l'ellipse
    :param hauteur: hauteur de l'ellipse
    :param couleur: couleur à utiliser pour remplir l'ellipse
    :returns: un graphique avec l'ellipse spécifiée
    """
    return en_ellipse(largeur, hauteur, couleur)


def secteur_circulaire(rayon: float, angle: float, couleur: Couleur) -> Graphique:
    """
    Crée un secteur circulaire appartenant à un cercle du rayon indiqué, rempli
    d'une couleur.

    Un secteur circulaire est une portion de cercle comprise entre deux rayons
    et un arc.
    Si l'on considère le cercle comme une horloge, le premier rayon "pointe"
    dans la direction de 3 heures. L'`angle` détermine la position du deuxième
    rayon, calculée à partir de la position du premier rayon dans le sens
    inverse des aiguilles d'une montre. Un angle de 360 degrés correspond à un
    cercle complet.

    Le point d'ancrage se trouve au centre du cercle à partir duquel le
    secteur circulaire est pris.

    :param rayon: rayon du cercle duquel est pris le secteur circulaire
    :param angle: angle au centre, en degrés
    :param couleur: couleur à utiliser pour remplir le secteur circulaire
    :returns: un graphique avec le secteur circulaire spécifié
    """
    return circular_sector(rayon, angle, couleur)


def triangle(cote1: float, cote2: float, angle: float, couleur: Couleur) -> Graphique:
    """
    Crée un triangle en spécifiant deux côtés et l'angle qui les sépare, remplis
    d'une couleur.
    Le premier côté s'étend horizontalement vers la droite. L'angle spécifie la
    rotation du deuxième côté, dans le sens inverse des aiguilles d'une montre,
    par rapport au premier.

    Pour tous les triangles, à l'exception des triangles obtus, le point
    inférieur gauche du graphique résultant coïncide avec le sommet du triangle
    dont l'angle a été spécifié.

    Le point d'ancrage est le centroïde du triangle.

    :param cote1: longueur du premier côté (horizontal) du triangle
    :param cote2: longueur du second côté du triangle
    :param angle: angle compris entre les deux côté, en degrés
    :param couleur: couleur à utiliser pour remplir le secteur circulaire
    :returns: un graphique avec le triangle spécifié
    """
    return en_triangle(cote1, cote2, angle, couleur)


def texte(contenu: str, police: str, points: float, couleur: Couleur) -> Graphique:
    """
    Crée un graphique avec le texte rendu à l'aide de la police, de la taille et
    de la couleur spécifiées.

    Lorsque la police True-Type indiquée n'est pas disponible dans le système,
    une police très basique toujours disponible est utilisée à la place. Le
    graphique qui en résulte a la taille minimale nécessaire pour contenir
    l'ensemble du texte.

    La point d'ancrage est alignée horizontalement sur la gauche et
    verticalement sur la ligne de base du texte.

    :param contenu: le texte à présenter
    :param police: le nom de la police (par exemple, "Arial" ou "Fira Sans")
    :param points: la taille en points typographiques (par exemple, 16)
    :param couleur: la couleur à utiliser pour le rendu du texte
    :returns: le texte spécifié sous forme de graphique
    """
    return text(contenu, police, points, couleur)
